import logging
import numpy 
from apgl.util import *
from apgl.graph import * 
from apgl.egograph.EgoGenerator import EgoGenerator 
from apgl.io import * 

class AbstractDiffusionSimulator():
    """
    An abstract class which learns from real survey data and then runs a simulation
    over a network to model information diffusion. 
    """

    def generateRandomGraph(self, egoFileName, alterFileName, numVertices, infoProb, graphType, p=0.1, neighbours=10):
        Parameter.checkInt(numVertices, 0, float('inf'))
        Parameter.checkFloat(infoProb, 0.0, 1.0)
        Parameter.checkString(graphType, ["SmallWorld", "ErdosRenyi"])
        Parameter.checkFloat(p, 0.0, 1.0)
        Parameter.checkInt(neighbours, 0, float('inf'))

        #Learn the distribution of the egos and alters
        eCsvReader = EgoCsvReader()

        self.egoQuestionIds = eCsvReader.getEgoQuestionIds()
        (X1, _) = eCsvReader.readFile(egoFileName, self.egoQuestionIds)

        alterQuestionIds = eCsvReader.getAlterQuestionIds()
        (X2, _) = eCsvReader.readFile(alterFileName, alterQuestionIds)

        X = numpy.r_[X1, X2]

        undirected = True

        egoGenerator = EgoGenerator()
        (mu, sigmaSq) = Util.computeMeanVar(X)
        vList = egoGenerator.generateIndicatorVertices2(numVertices, mu, sigmaSq, infoProb, X.min(0), X.max(0))

        self.graph = SparseGraph(vList, undirected)
        logging.info("Size of graph: " + str(self.graph.getNumVertices()))

        #Create the graph edges according to the small world model
        if graphType == "SmallWorld":
            graphGen = SmallWorldGenerator(self.graph)
            self.graph = graphGen.generateGraph(p, neighbours, self.edgeWeight)
        elif graphType == "ErdosRenyi":
            graphGen = ErdosRenyiGenerator(self.graph)
            self.graph = graphGen.generateGraph(p, self.edgeWeight)
        else:
            raise ValueError("Unknown graph type: " + graphType)

        logging.info("Number of edges of graph: " + str(self.graph.getNumEdges()))
        logging.info("Average number of edges per node: " + str(self.graph.getNumEdges()/float(numVertices)))
        #logging.info("Clustering coefficient: " + str(self.graph.clusteringCoefficient()))
        #logging.info("Degree distribution " + str(self.graph.degreeDistribution()))

        #This stuff is slow
        if numVertices <= 200:
            logging.info("Mean geodesic distance " + str(self.graph.geodesicDistance()))
            logging.info("Harmonic mean geodesic distance " + str(self.graph.harmonicGeodesicDistance()))

        return self.graph