
import numpy
import pickle
import logging 
from apgl.util import *
from apgl.graph.AbstractSingleGraph import AbstractSingleGraph
from apgl.graph.VertexList import VertexList

class AbstractMatrixGraph(AbstractSingleGraph):
    """
    An abstract graph object with single edges between vertices. The edge labels
    are stored in a weight matrix, and vertices are stored in a VertexList 
    """

    def getWeightMatrix(self):
        """
        Returns a numpy array of the weight matrix of this graph.
        """
        Util.abstract()

    def laplacianMatrix(self, outDegree=True):
        """
        Return the Laplacian matrix of this graph. For a directed graph the
        uses the out-degree
        """
        numVertices = self.getNumVertices()
        W = self.getWeightMatrix()

        if outDegree:
            return W + numpy.diag(numpy.dot(W, numpy.ones((numVertices, 1))).ravel())
        else:
            return W + numpy.diag(numpy.dot(numpy.ones((1, numVertices)), W).ravel())

    def getAllVertexIds(self):
        """
        Returns a list of all the vertex IDs of this graph. 
        """
        return range(0, self.vList.getNumVertices())

    def getVertexList(self):
        """
        Returns the VertexList object of this graph. 
        """
        return self.vList

    def getVertex(self, vertexIndex):
        """
        Returns the vertex associated with the given vertex index.

        :param vertexIndex: the index of the vertex.
        :type vertexIndex: :class:`int`
        """
        Parameter.checkIndex(vertexIndex, 0, self.vList.getNumVertices())
        return self.vList.getVertex(vertexIndex)

    def setVertex(self, vertexIndex, vertex):
        """
        Set the vertex with given index to a particular value. 

        :param vertexIndex: the index of the vertex.
        :type vertexIndex: :class:`int`

        :param vertex: the value of the vertex.
        :type vertex: :class:`numpy.array`
        """
        Parameter.checkIndex(vertexIndex, 0, self.vList.getNumVertices())
        self.vList.setVertex(vertexIndex, vertex)

    def getNumVertices(self):
        """
        Return the number of vertices in this graph. 
        """
        return self.vList.getNumVertices()

    def getNumVertexFeatures(self):
        """
        Return the number of vertex features.
        """
        return self.vList.getNumFeatures()

    def addEdge(self, vertexIndex1, vertexIndex2, edge=1):
        """
        Add a non-zero edge between two vertices.

        :param vertexIndex1: The index of the first vertex.
        :type vertexIndex1: :class:`int`

        :param vertexIndex2: The index of the second vertex.
        :type vertexIndex2: :class:`int`

        :param edge: The value of the edge.
        :type edge: :class:`float`
        """
        Parameter.checkIndex(vertexIndex1, 0, self.vList.getNumVertices())
        Parameter.checkIndex(vertexIndex2, 0, self.vList.getNumVertices())

        if edge == 0 or edge == float('inf'):
            raise ValueError("Cannot add a zero or infinite edge")

        self.W[vertexIndex1, vertexIndex2] = edge
        if self.undirected:
            self.W[vertexIndex2, vertexIndex1] = edge

    def addEdges(self, edgeIndexArray, edgeValues=[]):
        """
        Takes a numpy array of edge index pairs, and edge values and adds them
        to this graph. The array is 2 dimensional such that each row is a pair
        of edge indices.

        :param edgeIndexArray: The array of edge indices with each being a pair of indices.
        :type edgeIndexArray: :class:`numpy.ndarray`

        :param edgeValues: The list of edge values
        :type edgeValues: :class:`list`
        """
        if (edgeIndexArray < 0).any() or (edgeIndexArray >= self.vList.getNumVertices()).any():
            raise ValueError("Invalid indices for edges.")
        

        if edgeValues != []:
            if (edgeValues == 0).any():
                raise ValueError("Invalid entry, found zero edge value(s): " + str(numpy.nonzero(edgeValues==0)[0]))

            if self.undirected:
                for i in range(edgeIndexArray.shape[0]):
                    self.W[edgeIndexArray[i, 0], edgeIndexArray[i, 1]] = edgeValues[i]
                    self.W[edgeIndexArray[i, 1], edgeIndexArray[i, 0]] = edgeValues[i]
            else:
                for i in range(edgeIndexArray.shape[0]):
                    self.W[edgeIndexArray[i, 0], edgeIndexArray[i, 1]] = edgeValues[i]
        else:
            if self.undirected:
                for i in range(edgeIndexArray.shape[0]):
                    self.W[edgeIndexArray[i, 0], edgeIndexArray[i, 1]] = 1
                    self.W[edgeIndexArray[i, 1], edgeIndexArray[i, 0]] = 1
            else:
                for i in range(edgeIndexArray.shape[0]):
                    self.W[edgeIndexArray[i, 0], edgeIndexArray[i, 1]] = 1

    def getEdge(self, vertexIndex1, vertexIndex2):
        """
        Get the value of an edge.

        :param vertexIndex1: The index of the first vertex.
        :type vertexIndex1: :class:`int`

        :param vertexIndex2: The index of the second vertex.
        :type vertexIndex2: :class:`int`
        """
        Parameter.checkIndex(vertexIndex1, 0, self.vList.getNumVertices())
        Parameter.checkIndex(vertexIndex2, 0, self.vList.getNumVertices())

        if self.W[vertexIndex1, vertexIndex2]==0:
            return None
        else:
            return self.W[vertexIndex1, vertexIndex2]

    def getEdgeValues(self, edgeArray):
        """
        Take an array of n x 2 of vertex indices and return the corresponding edge
        values.
        """
        edgeValues = numpy.zeros(edgeArray.shape[0])

        for i in range(edgeValues.shape[0]):
            edgeValues[i] = self.getEdge(edgeArray[i, 0], edgeArray[i, 1])

        return edgeValues 

    def removeEdge(self, vertexIndex1, vertexIndex2):
        """
        Remove an edge between two vertices.

        :param vertexIndex1: The index of the first vertex.
        :type vertexIndex1: :class:`int`

        :param vertexIndex2: The index of the second vertex.
        :type vertexIndex2: :class:`int`
        """
        Parameter.checkIndex(vertexIndex1, 0, self.vList.getNumVertices())
        Parameter.checkIndex(vertexIndex2, 0, self.vList.getNumVertices())
        self.W[vertexIndex1, vertexIndex2] = 0
        if self.undirected:
            self.W[vertexIndex2, vertexIndex1] = 0

    def isUndirected(self):
        """
        Returns true if this graph is undirected otherwise false. 
        """
        return self.undirected

    def getAllEdges(self):
        """
        For an undirected graph will get v1>=v2. Return a matrix with 2 columns
        """
        (rows, cols) = numpy.nonzero(self.W)
        edges = numpy.c_[rows, cols]

        if self.undirected and edges.shape[0] != 0:
            edges = edges[edges[:,0] >= edges[:,1], :]

        return edges

    def removeAllEdges(self):
        """
        Removes all edges from this graph. 
        """
        self.W = self.W*0

    def __str__(self):
        output = str(type(self)) + ": "
        output = output + "no. vertices " + str(self.getNumVertices()) + ", no. edges, " + str(self.getNumEdges()) + " undirected: " + str(self.undirected)
        return output


    def diameter(self):
        """
        Finds the diameter of a graph i.e. the longest shortest path.
        """
        if self.getNumEdges() == 0: 
            return 0 

        P = self.floydWarshall(False)
        return int(numpy.max(P[P!=float('inf')]))

    def effectiveDiameter(self, q):
        """
        The effective diameter  is the minimum d such that for a fraction q of
        reachable node pairs, the path length is at most d. This is more rubust
        than the standard diamter method. 

        :param k: The fraction of node pairs to consider.
        :type k: :class:`float`
        """
        Parameter.checkFloat(q, 0.0, 1.0)

        if self.getNumEdges() == 0:
            return 0

        P = self.floydWarshall(False)
        paths = numpy.sort(P[P!=float('inf')])
        ind = numpy.floor(paths.shape[0]*q)-1
        return int(paths[ind])
        

    def floydWarshall(self, useWeights=True):
        """
        Use the Floyd-Warshall algorithm to find the shortest path between all pairs
        of vertices. If useWeights is true, then the weights are used to compute the
        path, otherwise adjacencies are used.

        :param useWeights: Whether to use the edge weight to compute path cost. 
        :type useWeights: :class:`bool`
        """
        numVertices = self.vList.getNumVertices()
        
        if useWeights:
            P = self.getWeightMatrix().copy()
        else:
            P = self.adjacencyMatrix().copy()
            
        P[P==0] = float('inf')
        j = numpy.ones(numVertices)

        for k in range(0, numVertices):
            P2 = numpy.outer(P[:, k], j) + numpy.outer(j, P[k, :])
            P = numpy.minimum(P, P2)

        return P

    def maxProductPaths(self):
        """
        Find the maximum product paths between all pairs of vertices using
        a modified version of the Floyd-Warshall algorithm.
        """
        numVertices = self.vList.getNumVertices()
        P = self.getWeightMatrix().copy()
        stepSize = 100

        for k in range(0, numVertices):
            Util.printIteration(k, stepSize, numVertices)
            P2 = numpy.outer(P[:, k], P[k, :])
            P = numpy.maximum(P, P2)

        return P

    def geodesicDistance(self):
        """
        Compute the mean geodesic distance for a graph. This is denoted by
        1/(1/2 n(n+1)) \sum_{i<=j} d_ij where d_ij is the shortest path length
        between i and j. Note that if i and j are not connected we assume a
        path length of 0.
        """
        n = self.getNumVertices()
        P = self.floydWarshall()
        P[P==numpy.inf] = 0
        P = P - numpy.diag(numpy.diag(P))
        distanceSum = numpy.sum(numpy.triu(P))

        return 2*distanceSum/(n*(n+1))

    def harmonicGeodesicDistance(self):
        """
        Compute the inverse "harmonic mean" geodesic distance for a graph. This is
        denoted by 1/(1/2 n(n+1)) \sum_{i<=j} d_ij^-1 where d_ij is the shortest
        path length between i and j. The distance from a node to itself is infinite. 
        """
        n = self.getNumVertices()
        P = self.floydWarshall()
        P = P + numpy.diag(numpy.ones(n)*numpy.inf)
        P = 1/P
        distanceSum = numpy.sum(numpy.triu(P))

        return (2*distanceSum)/(n*(n+1))

    def adjacencyMatrix(self):
        """
        Return the adjacency matrix in dense format. Warning: should not be used
        unless sufficient memory is available to store the dense matrix.
        """

        W2 = self.getWeightMatrix().copy()
        W2[W2.nonzero()] = 1
        return W2

    def maybeIsomorphicWith(self, graph):
        """
        Returns false if graph is definitely not isomorphic with the current graph,
        however a True may mean the graphs are not isomorphic. Makes a comparison
        with the eigenvalues of the Laplacian matrices. 
        """
        L1 = self.laplacianMatrix()
        L2 = graph.laplacianMatrix()

        sigma1, V1 = numpy.linalg.eig(L1)
        sigma2, V2 = numpy.linalg.eig(L2)

        tol = 10**-6

        return numpy.linalg.norm(sigma1 - sigma2) <= tol

    def complement(self):
        """
        Returns a graph with identical vertices (same reference) to the current one, but with the
        complement of the set of edges. Edges that do not exist have weight 1.
        """

        Util.abstract()

    def vertexLabelPairs(self, edgeArray):
        """
        Create an array of pairs of vertex labels for each edge in the graph. Returns
        a tuple of this array and a vector of corresponding edge weights, for the
        indices of the full edge index array as given by getAllEdges.

        :param edgeArray: A numpy array with 2 columns, and values corresponding to vertex indices.
        :type edgeArray: :class:`numpy.ndarray`

        """
        #Parameter.checkList(edgeArray, Parameter.checkIndex, (0, self.getNumVertices()))

        numFeatures = self.getVertexList().getNumFeatures()
        X = numpy.zeros((edgeArray.shape[0], numFeatures*2))

        for i in range(edgeArray.shape[0]):
            X[i, 0:numFeatures] = self.getVertex(edgeArray[i, 0])
            X[i, numFeatures:numFeatures*2] = self.getVertex(edgeArray[i, 1])

        return X 

    def save(self, filename):
        """
        Save the graph object to the corresponding file. The python pickle package
        is used for serialisation, and data is stored in a binary format.
        
        :param filename: The name of the file to save.
        :type filename: :class:`str`
        """
        Parameter.checkClass(filename, str) 

        file = open(filename, 'wb')
        pickle.dump(self, file)
        file.close()

        logging.info("Dumped file into " + filename)

    @staticmethod
    def load(filename):
        """
        Load the graph object from the corresponding file. The python pickle package
        is used for serialisation, and data is loaded in a binary format as created
        using save().

        :param filename: The name of the file to load.
        :type filename: :class:`str`
        """
        Parameter.checkClass(filename, str)

        file = open(filename, 'rb')
        graph = pickle.load(file)
        file.close()

        logging.info("Loaded pickled graph: " + filename + " with " + str(graph.getNumVertices()) + " vertices and " + str(graph.getNumEdges()) + " edges")

        return graph

    def setVertexList(self, vList):
        """
        Assign a new VertexList object to this graph. The number of vertices in the
        VertexList must be the same as in the graph. 
        """
        Parameter.checkClass(vList, VertexList)

        if vList.getNumVertices() != self.vList.getNumVertices():
            raise ValueError("Can only set to a VertexList with same number of vertices.")

        self.vList = vList 

    def setWeightMatrix(self, W):
        """
        Set the weight matrix of this graph. Requires as input an ndarray with the
        same dimensions as the current weight matrix. Edges are represented by
        non-zero edges.

        :param W: The name of the file to load.
        :type W: :class:`ndarray`
        """
        Parameter.checkClass(W, numpy.ndarray)

        if W.shape != (self.vList.getNumVertices(), self.vList.getNumVertices()):
            raise ValueError("Weight matrix has wrong shape : " + str(W.shape))

        if self.undirected and (W != W.T).any():
            raise ValueError("Weight matrix of undirected graph must be symmetric")


        self.W = W 


    vList = None
    undirected = None