'''
A class to model a simple directed graph with edge weights

Vertex indices are from 0 to numVertices-1 have no information about them 
Edges are a list of 3 items [from, to, weight]
'''

import numpy
from apgl.graph.AbstractMatrixGraph import AbstractMatrixGraph
from apgl.graph.AbstractVertexList import AbstractVertexList
from apgl.util.Parameter import Parameter

class DenseGraph(AbstractMatrixGraph):
    def __init__(self, vertexList, undirected=True):
        """
        Create a DenseGraph with a given AbstractVertexList, and specify whether directed.

        :param vList: the initial set of vertices as a AbstractVertexList object.
        :type vList: :class:`apgl.graph.AbstractVertexList`

        :param undirected: a boolean variable to indicate if the graph is undirected.
        :type undirected: :class:`boolean`
        """
        Parameter.checkClass(vertexList, AbstractVertexList)
        Parameter.checkBoolean(undirected)

        self.vList = vertexList
        self.W = numpy.zeros((vertexList.getNumVertices(), vertexList.getNumVertices()))
        self.undirected = undirected

    def getNumEdges(self):
        """
        Returns the total number of edges in this graph.
        """
        if self.undirected:
            return (numpy.flatnonzero(self.W).shape[0] + numpy.flatnonzero(numpy.diag(self.W)).shape[0])/2
        else: 
            return numpy.flatnonzero(self.W).shape[0]

    def getNumDirEdges(self):
        """
        Returns the number of edges, taking this graph as a directed graph. 
        """
        return numpy.flatnonzero(self.W).shape[0]
    
    def getWeightMatrix(self):
        """
        Return the weight matrix as a numpy array. 
        """
        return self.W

    def neighbours(self, vertexIndex):
        """
        Return an array of the indices of the neighbours of the given vertex.
        
        :param vertexIndex: the index of a vertex.
        :type vertexIndex: :class:`int`
        """
        Parameter.checkIndex(vertexIndex, 0, self.vList.getNumVertices())

        nonZeroIndices =  numpy.nonzero(self.W[vertexIndex, :])
        neighbourIndices = nonZeroIndices[0]
        
        return neighbourIndices

    def complement(self):
        """
        Returns a graph with identical vertices (same reference) to the current one, but with the
        complement of the set of edges. Edges that do not exist have weight 1.
        """

        graph = DenseGraph(self.vList, self.undirected)
        graph.W = numpy.array(self.W == 0, numpy.float64)
        return graph 
        
    undirected = None
    vList = None
    W = None
    