

from apgl.graph.GraphGenerator import GraphGenerator
from apgl.util.Parameter import Parameter 
import numpy

class SmallWorldGenerator(GraphGenerator):
    '''
    Generates a random graph using the Small World Model proposed by Watts and
    Strogatz. We start with a regular lattice and rewire nodes according to some
    probability p. Only works with undirected graphs currently.
    '''
    def __init__(self, graph):
        if not graph.isUndirected(): 
            raise ValueError
        self.graph = graph; 
    
        
    #
    def generateGraph(self, p, k, edgeWeight=1):
        '''
        Create a small-world graph with k initial neighbours and a re-wiring probability p

        :param p: the probability of rewiring an edge.
        :type p: :class:`float`

        :param k: the number of neighbours in the regular lattice.
        :type k: :class:`int`

        :param edgeWeight: the non-zero weight of newly created edges.
        :type edgeWeight: :class:`float`
        '''
        Parameter.checkFloat(p, 0.0, 1.0)
        Parameter.checkIndex(k, 0, self.graph.getNumVertices())

        numVertices = self.graph.getNumVertices()
        
        #First, create a lattice with k neighbours
        #For each vertex, create an edge to the next k vertices
        for i in range(0, numVertices): 
            for j in range(0, k): 
                self.graph.addEdge(i, (i+j+1)%numVertices, edgeWeight)
        
        #Now, let's get re-wiring, but only for forward links and no self-edges
        for i in range(0, numVertices): 
            v = numpy.random.rand(k, 1)
            for j in range(0, k):
                if (v[j] < p):  
                    ind = numpy.random.randint(0, numVertices)

                    if self.graph.getEdge(i, ind)==None and ind!=i:
                        self.graph.addEdge(i, ind, edgeWeight)
                        self.graph.removeEdge(i, (i+j+1)%numVertices)
                    
        return self.graph

    def clusteringCoefficient(self, p, k):
        '''
        Returns the clustering coefficient for the graph created with edge probability
        p, and number of neighbours k.
        '''
        Parameter.checkFloat(p, 0.0, 1.0)
        Parameter.checkIndex(k, 0, self.graph.getNumVertices())

        if k == 1:
            return 0
        else:
            return (float(3*(k-1))/(2*(2*k-1))) * (1-p)**3
     
    graph = None
        