
from apgl.graph.AbstractMatrixGraph import AbstractMatrixGraph
from apgl.graph.AbstractVertexList import AbstractVertexList
from apgl.util.Util import Util
from apgl.util.Parameter import Parameter
import scipy.sparse as sparse
import scipy
import numpy

class SparseGraph(AbstractMatrixGraph):
    '''
    Represents a graph, which can be directed or undirected, and has weights
    on the edges. Memory usage is efficient for sparse graphs. The list of vertices
    is immutable (see VertexList), however edges can be added or removed. Only
    non-zero edges can be added.
    '''
    def __init__(self, vList, undirected=True):
        """
        Create a SparseGraph with a given AbstractVertexList, and specify whether directed.

        :param vList: the initial set of vertices as a AbstractVertexList object.
        :type vList: :class:`apgl.graph.AbstractVertexList`
        
        :param undirected: a boolean variable to indicate if the graph is undirected.
        :type undirected: :class:`boolean`
        """
        Parameter.checkClass(vList, AbstractVertexList)
        Parameter.checkBoolean(undirected)

        self.vList = vList
        self.undirected = undirected
        
        self.W = sparse.csr_matrix((vList.getNumVertices(), vList.getNumVertices()))
        
    def neighbours(self, vertexIndex):
        """
        Return an array of the indices of neighbours.

        :param vertexIndex: the index of a vertex.
        :type vertexIndex: :class:`int`
        """
        Parameter.checkIndex(vertexIndex, 0, self.vList.getNumVertices())
        nonZeroInds = numpy.nonzero(self.W.tocsr()[vertexIndex, :])
        neighbours = nonZeroInds[1]
        
        return neighbours
    
    def getNumEdges(self):
        """
        Returns the total number of edges in this graph. 
        """
        #Note that self.W.getnnz() doesn't seem to work correctly 
        if self.undirected == True:
            return (self.W.nonzero()[0].shape[0] + Util.diag(self.W).getnnz())/2
        else: 
            return self.W.nonzero()[0].shape[0]

    def getNumDirEdges(self):
        """
        Returns the number of edges, taking this graph as a directed graph.
        """
        return self.W.nonzero()[0].shape[0]
    
    def clusteringCoefficient(self):
        """ 
        Find the clustering coefficient of this graph. 
        """
        A = self.W / self.W
        A2 = numpy.dot(A, A)
        A3 = numpy.dot(A2, A)
        
        diagSumA2 = 0 
        
        for i in range(0, A2.shape[0]): 
            diagSumA2 = diagSumA2 + A2[i, i]
        
        numTriples = A2.sum()-diagSumA2
        
        diagSumA3 = 0 
        
        for i in range(0, A3.shape[0]): 
            diagSumA3 = diagSumA3 + A3[i, i]
        
        if numTriples == 0: 
            return 0 
        else:    
            return diagSumA3/numTriples
    
    def degreeDistribution(self):
        """ 
        Return a vector of (out)degree distributions. The ith element of the vector
        corresponds to the frequency of degree i.
        """ 
        degrees = self.outDegreeSequence()
        
        binCounts = numpy.bincount(degrees) 
        return binCounts

    def outDegreeSequence(self):
        """
        Return a vector of the (out)degree sequence for each vertex.
        """
        degrees = numpy.zeros(self.W.shape[0], dtype=numpy.int32)

        for i in range(0, self.W.shape[0]):
            degrees[i] = self.W[i, :].getnnz()

        return degrees 

    def inDegreeDistribution(self):
        """
        Returns a vector of in-degree distributions. The ith element of the vector
        corresponds to the frequency of degree i. 
        """
        degrees = self.inDegreeSequence()

        binCounts = numpy.bincount(degrees)
        return binCounts

    def inDegreeSequence(self):
        """
        Return a vector of the (in)degree sequence for each vertex.
        """
        degrees = numpy.zeros(self.W.shape[0], dtype=numpy.int32)

        for i in range(0, self.W.shape[0]):
            degrees[i] = self.W[:, i].getnnz()

        return degrees 
    
    def dijkstrasAlgorithm(self, vertexIndex):
        """ 
        Run Dijkstras Algorithm on the graph for a given source vertex. We ignore weights
        and simply use distance as the (unweighted) path length.
        
        :param vertexIndex: the index of the source vertex. 
        :type vertexIndex: :class:`int`
        """
        Parameter.checkIndex(vertexIndex, 0, self.vList.getNumVertices())

        distance = numpy.ones(self.W.shape[0])*numpy.inf
        previous = numpy.ones(self.W.shape[0])*-1
        distance[vertexIndex] = 0
        
        A = self.W / self.W
        vertexIndices = numpy.array(range(0, self.W.shape[0]))
        
        while vertexIndices.shape[0] != 0: 
            minVertexIndex = scipy.argmin(distance[vertexIndices])          
            minVertexDistance = distance[vertexIndices][minVertexIndex]
            
            if  minVertexDistance == numpy.inf: 
                break 
            
            (rows, cols) = A[vertexIndices[minVertexIndex], :].nonzero()
            
            for i in cols: 
                newDistance = minVertexDistance + A[vertexIndices[minVertexIndex], i]
                if newDistance < distance[i]: 
                    distance[i] =  newDistance
                    previous[i] = minVertexIndex
                
            vertexIndices = numpy.delete(vertexIndices, minVertexIndex)
                    
        return distance

    def subgraph(self, vertexIndices):
        """
        Pass in a list or set of vertexIndices and returns the subgraph containing
        those vertices only, and edges between them.

        :param vertexIndex: the indices of the subgraph vertex.
        :type vertexIndex: :class:`list`
        """
        vertexIndices = list(set(vertexIndices))
        vList = self.vList.subList(vertexIndices)

        subGraph = SparseGraph(vList, self.undirected)

        subGraph.W = self.W[vertexIndices, :]
        subGraph.W = subGraph.W[:, vertexIndices]

        return subGraph

    def getWeightMatrix(self):
        """
        Return the weight matrix in dense format. Warning: should not be used
        unless sufficient memory is available to store the dense matrix. 
        """

        return self.W.toarray()

    def getSparseWeightMatrix(self):
        """
        Returns the original sparse weight matrix.
        """

        return self.W

    def edgeAddition(self, sGraph):
        """
        Add the edge weights of the input graph to the current one. Results in a
        union of the edges.

        :param sGraph: the input graph.
        :type sGraph: :class:`apgl.graph.SparseGraph`
        """
        if sGraph.getNumVertices() != self.getNumVertices():
            raise ValueError("Can only add edges from graph with same number of vertices")

        #The ideal way is to add both weight matrices together, but this results in a csr
        #We'll just do this manually
        nonZeros = numpy.nonzero(sGraph.W)

        for i in range(len(nonZeros[0])):
            ind1 = nonZeros[0][i]
            ind2 = nonZeros[1][i]
            self.W[ind1, ind2] = self.W[ind1, ind2] +  sGraph.W[ind1, ind2]
            
    def copy(self):
        """
        Returns a copy of this object, which also has a copy of the VertexList.
        """
        sGraph = SparseGraph(self.vList.copy(), self.undirected)
        sGraph.W = self.W.copy()
        return sGraph

    def egoGraph(self, vertexIndex):
        """
        Returns the subgraph composed of the given vertex and its immediate neighbours.
        In the new graph, the ego is index 0 and neighbours are indexed in order
        after 0.

        :param vertexIndex: the index of the source vertex.
        :type vertexIndex: :class:`int`
        """
        Parameter.checkIndex(vertexIndex, 0, self.vList.getNumVertices())

        neighbours = self.neighbours(vertexIndex)
        egoGraphIndices = numpy.r_[numpy.array([vertexIndex]), neighbours]
        return self.subgraph(egoGraphIndices)

    def complement(self):
        """
        Returns a graph with identical vertices (same reference) to the current
        one, but with the complement of the set of edges. Edges that do not exist
        have weight 1. This makes a sparse graph dense.
        """

        graph = SparseGraph(self.vList, self.undirected)
        graph.W = sparse.csr_matrix(numpy.ones((self.vList.getNumVertices(), self.vList.getNumVertices())))

        A = self.W / self.W
        graph.W = graph.W - A 

        return graph

    def findTrees(self):
        """
        Returns a list of trees for a directed graph. The reason for only allowing
        directed graphs is that the root of a tree in an undirected graph is ambiguous.
        Each tree is represented by an list of indices of vertices in the graph. 
        """

        if self.isUndirected():
            raise ValueError("Can only find trees on directed graphs")

        trees = []

        #Find all vertices with in-degree 0
        for i in range(0, self.getNumVertices()):
            if self.W[:, i].getnnz() == 0:
                trees.append(self.depthFirstSearch(i))

        return trees

    def setWeightMatrix(self, W):
        """
        Set the weight matrix of this graph. Requires as input an ndarray with the
        same dimensions as the current weight matrix. Edges are represented by
        non-zero edges.

        :param W: The name of the file to load.
        :type W: :class:`ndarray`
        """
        Parameter.checkClass(W, numpy.ndarray)

        if W.shape != (self.vList.getNumVertices(), self.vList.getNumVertices()):
            raise ValueError("Weight matrix has wrong shape : " + str(W.shape))

        if self.undirected and (W != W.T).any():
            raise ValueError("Weight matrix of undirected graph must be symmetric")

        self.W = sparse.csr_matrix(W)

    #Class data 
    W = None
    vList = None
    undirected = None
    