from apgl.graph import * 
import unittest

import numpy 

class AbstractSingleGraphTest(unittest.TestCase):
    def setUp(self):
        self.numVertices = 10
        self.numFeatures = 3
        self.vList = VertexList(self.numVertices, self.numFeatures)

        self.dGraph = DenseGraph(self.vList)
        self.dGraph.addEdge(5, 1, 4)
        self.dGraph.addEdge(5, 2, 2)
        self.dGraph.addEdge(2, 7, 4)
        self.dGraph.addEdge(1, 9, 6)

        self.dGraph2 = DenseGraph(self.vList, False)
        self.dGraph2.addEdge(5, 1, 4)
        self.dGraph2.addEdge(5, 2, 2)
        self.dGraph2.addEdge(2, 7, 4)
        self.dGraph2.addEdge(1, 9, 6)

    def testToNetworkXGraph(self):
        try:
            import networkx
        except ImportError:
            pass 

        networkXGraph = self.dGraph.toNetworkXGraph()

        self.assertEquals(networkXGraph.get_edge_data(5, 1), {'value' : 4.0})
        self.assertEquals(networkXGraph.get_edge_data(5, 2), {'value' : 2.0})
        self.assertEquals(networkXGraph.get_edge_data(2, 7), {'value' : 4.0})
        self.assertEquals(networkXGraph.get_edge_data(1, 9), {'value' : 6.0})
        self.assertEquals(networkXGraph.get_edge_data(9, 1), {'value' : 6.0})

        vertexIndexList = []

        for i in networkXGraph.__iter__():
            vertexIndexList.append(i)

        vertexIndexList.sort() 
        self.assertTrue(vertexIndexList == range(self.numVertices))
        self.assertTrue(networkXGraph.edges() == [(1, 9), (1, 5), (2, 5), (2, 7)])
        self.assertTrue(type(networkXGraph) == networkx.Graph)

        #Now we test the case where we have a directed graph 
        networkXGraph = self.dGraph2.toNetworkXGraph()

        self.assertEquals(networkXGraph.get_edge_data(5, 1), {'value' : 4.0})
        self.assertEquals(networkXGraph.get_edge_data(5, 2), {'value' : 2.0})
        self.assertEquals(networkXGraph.get_edge_data(2, 7), {'value' : 4.0})
        self.assertEquals(networkXGraph.get_edge_data(1, 9), {'value' : 6.0})
        self.assertFalse(networkXGraph.has_edge(9, 1))

        vertexIndexList = []

        for i in networkXGraph.__iter__():
            vertexIndexList.append(i)

        vertexIndexList.sort()
        self.assertTrue(vertexIndexList == range(self.numVertices))
        self.assertTrue(networkXGraph.edges() == [(1, 9), (2, 7), (5, 1), (5, 2)])
        self.assertTrue(type(networkXGraph) == networkx.DiGraph)

        #Test a graph with no edges
        numVertices = 10
        numFeatures = 3
        vList = VertexList(numVertices, numFeatures)
        vList.setVertices(numpy.random.rand(numVertices, numFeatures))

        dGraph = DenseGraph(vList)
        networkXGraph = dGraph.toNetworkXGraph()

        self.assertTrue(networkXGraph.order() == numVertices)
        self.assertTrue(networkXGraph.size() == 0)

        self.assertTrue((networkXGraph.nodes(data=True)[0][1]['label'] == dGraph.getVertex(0)).all())


    def testDepthFirstSearch(self):
        numVertices = 10
        numFeatures = 0
        vList = VertexList(numVertices, numFeatures)

        dGraph = DenseGraph(vList)
        dGraph.addEdge(0, 1)
        dGraph.addEdge(1, 2)
        dGraph.addEdge(1, 3)
        dGraph.addEdge(2, 6)
        dGraph.addEdge(4, 5)

        self.assertEquals(dGraph.depthFirstSearch(0), set([0,1,2,3,6]))
        self.assertEquals(dGraph.depthFirstSearch(1), set([0,1,2,3,6]))
        self.assertEquals(dGraph.depthFirstSearch(6), set([0,1,2,3,6]))
        self.assertEquals(dGraph.depthFirstSearch(4), set([4, 5]))
        self.assertEquals(dGraph.depthFirstSearch(5), set([4, 5]))
        self.assertEquals(dGraph.depthFirstSearch(7), set([7]))

    def testFindComponents(self):
        numVertices = 10
        numFeatures = 0
        vList = VertexList(numVertices, numFeatures)

        dGraph = DenseGraph(vList)
        dGraph.addEdge(0, 1)
        dGraph.addEdge(1, 2)
        dGraph.addEdge(1, 3)
        dGraph.addEdge(2, 6)
        dGraph.addEdge(4, 5)

        self.assertEquals(dGraph.findConnectedComponents()[0], set([0,1,2,3,6]))
        self.assertEquals(dGraph.findConnectedComponents()[1], set([4, 5]))

        dGraph = DenseGraph(vList, False)
        self.assertRaises(ValueError, dGraph.findConnectedComponents)

    def testAddEdges(self):
        numVertices = 10
        numFeatures = 0
        vList = VertexList(numVertices, numFeatures)

        dGraph = DenseGraph(vList)

        edges = numpy.array([[0, 1], [1,2], [5,6], [3,1]])
        dGraph.addEdges(edges)

        self.assertTrue(dGraph.getEdge(0, 1) == 1)
        self.assertTrue(dGraph.getEdge(1, 2) == 1)
        self.assertTrue(dGraph.getEdge(5, 6) == 1)
        self.assertTrue(dGraph.getEdge(3, 1) == 1)
        self.assertTrue(dGraph.getNumEdges() == 4 )

        #Now try adding edge values

        dGraph = DenseGraph(vList)

        edges = numpy.array([[0, 1], [1,2], [5,6], [3,1]])
        edgeValues = numpy.array([23, 1.2, 8, 12])
        dGraph.addEdges(edges, edgeValues)

        self.assertTrue(dGraph.getEdge(0, 1) == 23)
        self.assertTrue(dGraph.getEdge(1, 2) == 1.2)
        self.assertTrue(dGraph.getEdge(5, 6) == 8)
        self.assertTrue(dGraph.getEdge(3, 1) == 12)
        self.assertTrue(dGraph.getNumEdges() == 4) 

    def testSetVertices(self):
        graph = DictGraph()

        vertexIndices = [1, 2, 3]
        vertices = ["a", "b", "c"]

        graph.setVertices(vertexIndices, vertices)

        vertexIndices2 = graph.getAllVertexIds()
        vertices2 = graph.getVertices(vertexIndices2)

        self.assertEquals(vertexIndices, vertexIndices2)
        self.assertEquals(vertices, vertices2)



if __name__ == '__main__':

    unittest.main()

