'''
Created on 31 Jul 2009

@author: charanpal
'''

import numpy
import numpy.random as rand
import logging
import inspect
import scipy.sparse as sparse 

class Util(object):
    '''
    A class with some general useful function that don't fit in anywhere else. Not very OO unfortunately. 
    '''
    def __init__(self):
        '''
        Constructor
        '''
        pass
    
    @staticmethod   
    def histogram(v):
        """
        Compute a histogram based on all unique elements in vector v 
        """ 
        if v.ndim != 1: 
            raise ValueError("Input must be a dimension 1 vector")
        
        uniqElements = numpy.unique(v)
        numElements = uniqElements.shape[0]
        hist = numpy.zeros(numElements)
        
        for i in range(0, numElements): 
            hist[i] = sum(v == uniqElements[i])

        return (hist, uniqElements)

    @staticmethod
    def mode(v):
        """
        Returns the mode of a 1D vectors, and the 1st more frequent element if more than 1
        """
        if v.ndim != 1:
            raise ValueError("Input must be a dimension 1 vector")

        uniqElements = numpy.unique(v)
        numElements = uniqElements.shape[0]
        maxFreq = 0
        mode = 0

        for i in range(0, numElements):
            freq = sum(v == uniqElements[i])

            if freq > maxFreq:
                maxFreq = freq
                mode = uniqElements[i]

        return mode
    
    @staticmethod  
    def sampleWithoutReplacement(sampleSize, totalSize):
        """ 
        Create a list of integers from 0 to totalSize, and take a random sample of size sampleSize. The 
        sample ordered. 
        """
        perm = rand.permutation(totalSize)
        perm = perm[0:sampleSize]
        perm = numpy.sort(perm)
        
        return perm 
    
    @staticmethod
    def randNormalInt(mean, sd, min, max):
        """
        Returns a normally distributed integer within a range (inclusive of min, max) 
        """
        i = round(rand.normal(mean, sd)); 
        
        while i<min or i>max: 
            i = round(random.normal(mean, sd)); 
            
        return i

    @staticmethod 
    def computeMeanVar(X):
        mu = numpy.mean(X, 0)
        X2 = X - mu
        sigma = numpy.dot(X2.T, X2)/X.shape[0]

        return (mu, sigma)

    @staticmethod 
    def printIteration(i, step, maxIter):
        if maxIter == 1:
            logging.info("Iteration: " + str(i) + " (1.0)")
        elif i % step == 0:
            #frm = inspect.stack()[1]
            #mod = inspect.getmodule(frm[0])
            #logging.info(mod.__name__ +  ": " + str(i) + " (" + str(float(i)/maxIter) + ")")
            logging.info("Iteration: " + str(i) + " (" + str("%.3f" % (float(i)/(maxIter-1))) + ")")
        elif i == maxIter-1:
            logging.info("Iteration: " + str(i) + " (" + str("%.3f" % (float(i)/(maxIter-1))) + ")")
            

    @staticmethod
    def diag(X):
        """
        Find the diagonal of a lil_matrix
        """
        d = sparse.lil_matrix((X.shape[0], 1))

        for i in range(0, X.shape[0]):
            d[i, 0] = X[i, i]

        return d 

    @staticmethod
    def countFrequencies(itemArray, numElements):
        """
        Takes a 1D numpy array and a max number of elements and returns a histogram of
        the frequency of each element from 0 to numElements-1
        """
        frequencies = numpy.zeros(numElements)

        for i in range(0, numElements):
            frequencies[i] = numpy.nonzero(itemArray == i)[0].shape[0]

        return frequencies

    @staticmethod
    def abstract():
        """
        This is a method to be put in abstract methods so that they are identified
        as such when called. 
        """
        import inspect
        caller = inspect.getouterframes(inspect.currentframe())[1][3]
        raise NotImplementedError(caller + ' must be implemented in subclass')


    @staticmethod
    def mdot(*args):
        """
        The optimisation method to find the quickest way of evaluating the matrix
        product is simple: we look for the pair which results in the smallest
        resulting matrix first.
        NOTE: When doing a multiplications as follows aa'X where a is a vector and
        X is a matrix, be careful to ensure a is 2D otherwise a'aX may result!!!! 
        """        
        if len(args) == 1:
            return args[0]
        elif len(args) == 2:
            for i in range(len(args)):
                logging.debug("Argument " + str(i) + " : " + str(args[i].shape))
            return numpy.dot(args[0], args[1])
        else:
            resultSizes = numpy.zeros(len(args)-1)

            for i in range(len(args)-1):
                if args[i].ndim == 2 and args[i+1].ndim == 2:
                    resultSizes[i] = args[i].shape[0]*args[i+1].shape[1]
                elif args[i].ndim > 2 or args[i+1].ndim > 2:
                    raise ValueError("mdot input arrays must be of dimension 2 or less")
                else:
                    resultSizes[i] = args[i].size * args[i+1].size

            j = numpy.argmin(resultSizes)
            newArgs = args[0:j] + (Util.mdot(args[j], args[j+1]), ) + args[j+2:]
            return Util.mdot(*newArgs)

    @staticmethod
    def rank(A, tol=1e-8):
        """
        Kindly borrowed from the following forum thread: 
        http://mail.scipy.org/pipermail/numpy-discussion/2008-February/031218.html
        """
        s = numpy.linalg.svd(A, compute_uv=False)
        return numpy.sum(numpy.where(s>tol, 1, 0))

    @staticmethod
    def randomChoice(v):
        """
        Make a random choice from a vector v of values which are unnormalised
        probabilities. Return the corresponding index.
        """
        cumV = numpy.cumsum(v)
        maxV = cumV[v.shape[0]-1]

        p = numpy.random.rand()*maxV
        return numpy.searchsorted(cumV, p)

