"""
A class to represent someone in the HIV epidemic network
"""

import numpy 
from apgl.util import * 

class HIVIndividual():
    def __init__(self):
        self.numFeatures = 7

        self.dobIndex = 0
        self.genderIndex = 1
        self.orientationIndex = 2

        #Time varying features 
        self.freeDegreeIndex = 3
        self.stateIndex = 4
        self.infectionTimeIndex = 5
        self.detectionTimeIndex = 6

        self.x = numpy.zeros(self.numFeatures)

        #Let's randomly select the characteristics
        self.setDateOfBirth(numpy.random.rand())
        self.setGender(Util.randomChoice(numpy.array([1, 1])))
        if self.getGender() == 0:
            self.setOrientation(0)
        else:
            self.setOrientation(Util.randomChoice(numpy.array([4, 1])))

        self.setState(0)

    def setDateOfBirth(self, dob):
        Parameter.checkFloat(dob, 0.0, float('inf'))
        self.x[self.dobIndex] = dob

    def getDateOfBirth(self):
        return self.x[self.dobIndex]

    def getAge(self, time):
        Parameter.checkFloat(time, self.x[self.dobIndex], float('inf'))
        return time - self.x[self.dobIndex]

    def setGender(self, gender):
        """
        Sets the gender - 0 = female and 1 = male. 
        """
        if gender not in [0, 1]:
            raise ValueError("Invalid gender : " + str(gender))

        self.x[self.genderIndex] = gender

    def getGender(self):
        return int(self.x[self.genderIndex])

    def setOrientation(self, orientation):
        """
        Sets the orientation where 0 = Hetro, 1 = Bisexual. All women are hetro
        apparently.
        """
        if orientation not in [0, 1]:
            raise ValueError("Invalid orientation : " + str(orientation))

        if orientation == 1 and self.getGender()==0:
            raise ValueError("Women cannot be bisexual")

        self.x[self.orientationIndex] = orientation

    def getOrientation(self):
        return int(self.x[self.orientationIndex])

    def setFreeDegree(self, freeDegree):
        Parameter.checkInt(freeDegree, 0, float('inf'))

        self.x[self.freeDegreeIndex] = freeDegree

    def getFreeDegree(self):
        return int(self.x[self.freeDegreeIndex]) 

    def setState(self, state):
        """
        Sets the state of the individual 0 = susceptible, 1 = Infected, 2 = Removed
        """
        if state not in [0, 1, 2]:
             raise ValueError("Invalid state : " + str(state))
         
        self.x[self.stateIndex] = state

    def getState(self):
        return int(self.x[self.stateIndex])

    def getInfectionTime(self):
        return self.x[self.infectionTimeIndex]

    def getDetectionTime(self):
        return self.x[self.detectionTimeIndex]

    def setInfected(self, time):
        Parameter.checkFloat(time, 0.0, float('inf'))
        
        if self.x[self.freeDegreeIndex] == 0:
            raise ValueError("Free degree of vertex is zero - cannot infect")

        if self.getState() == 1:
            raise ValueError("Person is already infected")
        
        self.setState(1)
        self.setFreeDegree(self.x[self.freeDegreeIndex]-1)
        self.x[self.infectionTimeIndex] = time

    def setDetected(self, time):
        """
        Corresponds to a detection of an individual
        """
        Parameter.checkFloat(time, 0.0, float('inf'))
        
        if self.getState() != 1:
            raise ValueError("Person must be infected to be detected")

        self.setState(2)
        self.x[self.detectionTimeIndex] = time

    def __str__(self):
        output = "DoB: " + str(self.getDateOfBirth()) + " Gender: " + str(self.getGender()) + " Orientation: " + str(self.getOrientation())
        output += " Free degree: " + str(self.getFreeDegree()) + " State: " + str(self.getState())
        return output